use crate::bundle::{
  poller::{BundlePoller, PollResult, PollerError, Result},
  Bundle,
};

use async_trait::async_trait;
use rusoto_core::RusotoError;
use rusoto_s3::{
  GetObjectError, GetObjectRequest, HeadObjectError, HeadObjectRequest, S3Client, S3,
};
use std::path::PathBuf;
use thiserror::Error;
use tokio_tar::Archive;

#[derive(Debug, Error)]
pub enum InternalError {
  #[error(transparent)]
  S3HeadObjectError {
    #[from]
    source: RusotoError<HeadObjectError>,
  },
  #[error(transparent)]
  S3GetObjectError {
    #[from]
    source: RusotoError<GetObjectError>,
  },
  #[error("Missing body")]
  S3MissingBody,
  #[error("Missing eTag")]
  S3MissingETag,
}

impl From<InternalError> for PollerError {
  fn from(err: InternalError) -> Self {
    PollerError::InternalPollerError {
      source: Box::new(err),
    }
  }
}

pub struct S3BundlePoller {
  client: S3Client,
  bucket: String,
  object_name: String,
}

impl S3BundlePoller {
  pub fn new(
    access_key: String,
    secret_key: String,
    endpoint: String,
    bucket: String,
    region: String,
    object_name: String,
  ) -> S3BundlePoller {
    let region = rusoto_core::Region::Custom {
      name: region,
      endpoint,
    };
    let credentials = rusoto_credential::StaticProvider::new_minimal(access_key, secret_key);
    let dispatcher = rusoto_core::HttpClient::new().unwrap();
    let client = S3Client::new_with(dispatcher, credentials, region);

    S3BundlePoller {
      client,
      bucket,
      object_name,
    }
  }
}

#[async_trait]
impl BundlePoller for S3BundlePoller {
  async fn poll(&self, active_bundle: &Option<Bundle>) -> Result<PollResult> {
    if let Some(active_bundle) = active_bundle {
      if active_bundle.etag.is_some() {
        let head_request = HeadObjectRequest {
          bucket: self.bucket.clone(),
          key: self.object_name.clone(),
          ..Default::default()
        };

        let head_response = self
          .client
          .head_object(head_request)
          .await
          .map_err(|source| InternalError::S3HeadObjectError { source })?;

        if head_response.e_tag == active_bundle.etag {
          log::info!(
            "S3BundlePoller: No updates found for object: {}",
            self.object_name
          );

          // Skip update download
          return Ok(PollResult::Skip);
        } else {
          log::info!(
            "S3BundlePoller: Object {} has changed. Starting bundle update.",
            self.object_name
          );
        }
      } else {
        log::error!("S3BundlePoller: Expected the active bundle to have a valid ETag.");

        return Err(PollerError::MissingBundleID);
      }
    }

    // Head object
    let head_request = HeadObjectRequest {
      bucket: self.bucket.clone(),
      key: self.object_name.clone(),
      ..Default::default()
    };

    let head_response = self
      .client
      .head_object(head_request)
      .await
      .map_err(|source| InternalError::S3HeadObjectError { source })?;

    Ok(PollResult::UpdateReady {
      etag: head_response.e_tag.ok_or(InternalError::S3MissingETag)?,
    })
  }

  async fn retrieve(&self, bundle_id: &str, path: PathBuf) -> Result<()> {
    log::info!("S3BundlePoller: Starting bundle download...");

    let get_object_request = GetObjectRequest {
      bucket: self.bucket.clone(),
      key: self.object_name.clone(),
      ..Default::default()
    };

    let get_object_response = self
      .client
      .get_object(get_object_request)
      .await
      .map_err(|source| InternalError::S3GetObjectError { source })?;

    let current_bundle_id = get_object_response
      .e_tag
      .ok_or(InternalError::S3MissingETag)?;

    if current_bundle_id != bundle_id {
      return Err(PollerError::MismatchedBundleID {
        original_id: String::from(bundle_id),
        current_id: current_bundle_id,
      });
    }

    log::info!("S3BundlePoller: Unpacking bundle...");

    let stream = get_object_response
      .body
      .ok_or(InternalError::S3MissingBody)?
      .into_async_read();

    // TODO: Write temp file to disk.
    let mut archive = Archive::new(stream);

    archive
      .unpack(path)
      .await
      .map_err(|err| PollerError::UnpackError { source: err })?;

    Ok(())
  }
}
