use actix::System;
use collective::thread::{
  self,
  monitor::{self, ThreadMonitor},
};
use espresso::{
  bundle::{self, Unbundler},
  config::Config,
  server::{self, Server},
  stats::{self, StatsServer},
};
use lazy_static::lazy_static;
use std::{
  collections::HashSet,
  sync::{mpsc, Arc},
};
use thiserror::Error;
use tokio::sync::RwLock;

lazy_static! {
  static ref MONITOR: ThreadMonitor = ThreadMonitor::new();
}

#[derive(Debug, Error)]
pub enum ServeError {
  #[error(transparent)]
  Unbundle {
    #[from]
    source: bundle::BundleError,
  },
  #[error(transparent)]
  ServeError {
    #[from]
    source: Box<server::ServerError>,
  },
  #[error(transparent)]
  ServeStats {
    #[from]
    source: stats::StatsError,
  },
  #[error(transparent)]
  MonitorError {
    #[from]
    source: monitor::Error,
  },
  #[error("Issue with receiving notification")]
  RecvNotify,
}

type Result<T, E = ServeError> = std::result::Result<T, E>;

pub async fn serve(config: Arc<Config>) -> Result<()> {
  MONITOR.init()?;

  // Set up a channel for receiving thread notifications.
  let (monitor_tx, monitor_rx) = mpsc::channel();

  // Keep track of what threads have been started.
  let mut server_thread_ids = HashSet::new();

  // Set up unbundler.
  let serve_dir = Arc::new(RwLock::new(None));
  let unbundler = Arc::new(Unbundler::new(config.clone(), serve_dir.clone()));

  // Set up main server.
  let server = Server::new(config.server.clone(), serve_dir);

  let (server_handle, server_thread_handle) =
    server
      .spawn(monitor_tx.clone())
      .await
      .map_err(|err| ServeError::ServeError {
        source: Box::new(err),
      })?;

  server_thread_ids.insert(server_thread_handle.thread().id());

  // Set up optional stats server.
  let mut maybe_stats_server_handle = None;

  match &config.stats {
    Some(stats_config) => {
      let stats_server = StatsServer::new(stats_config.clone(), unbundler.clone());

      let (stats_server_handle, stats_thread_handle) =
        stats_server.spawn(monitor_tx.clone()).await?;

      maybe_stats_server_handle = Some(stats_server_handle);
      server_thread_ids.insert(stats_thread_handle.thread().id());
    }
    None => {}
  }

  let unbundler_thread_handle = thread::handle::spawn(monitor_tx.clone(), move || {
    let sys = System::new();

    let result = sys.block_on(async move { unbundler.enter().await });

    if let Err(e) = result {
      log::error!("Unbundler failed: {:?}", e);
    }
  });
  let unbundler_thread_id = unbundler_thread_handle.thread().id();

  let monitor_thread_handle = thread::handle::spawn(monitor_tx.clone(), move || {
    let mut watched_thread_ids = HashSet::new();

    watched_thread_ids.insert(unbundler_thread_id);

    for server_thread_id in server_thread_ids {
      watched_thread_ids.insert(server_thread_id);
    }

    if MONITOR.watch(Some(&watched_thread_ids)).is_err() {
      log::error!("Failed to watch threads for panics.");
    }
  });

  // Wait for a thread to finish.
  loop {
    monitor_rx.recv().map_err(|_| ServeError::RecvNotify)?;

    if Ok(true) == monitor_thread_handle.get_end_handle().has_ended() {
      log::info!("Stopping servers due to a panic.");

      break;
    } else if Ok(true) == unbundler_thread_handle.get_end_handle().has_ended() {
      log::info!("Stopping servers due to unbundler shutdown.");

      break;
    }
  }

  // Stop server threads.
  server_handle.stop(true).await;

  if let Some(stats_server_handle) = maybe_stats_server_handle {
    stats_server_handle.stop(true).await;
  }

  Ok(())
}
