use std::path::{Path, PathBuf};
use thiserror::Error;

#[derive(Debug, Error)]
pub enum PathError {
  #[error(transparent)]
  StripPrefix {
    #[from]
    source: std::path::StripPrefixError,
  },
  #[error("Path has no parent")]
  NoParent,
  #[error("Unable to canonicalize path (Path: {}, Source: {})", path.display(), source)]
  Canonicalize {
    path: PathBuf,
    #[source]
    source: std::io::Error,
  },
}

/// Resolves a path relative to the current path or to the base path.
///
/// Note: This function does not check that the path resolves somewhere outside
/// of the base path.
pub fn resolve_path_within_tree(
  base_path: &Path,
  current_path: &Path,
  path: &Path,
) -> Result<PathBuf, PathError> {
  let new_path = if path.is_absolute() {
    base_path.join(path.strip_prefix("/")?)
  } else if current_path.is_dir() {
    current_path.join(path)
  } else {
    current_path
      .parent()
      .map(|x| x.join(path))
      .ok_or(PathError::NoParent)?
  };

  new_path
    .canonicalize()
    .map_err(|source| PathError::Canonicalize {
      path: new_path.clone(),
      source,
    })
}
