use actix_web::{
  dev::{Payload, ServiceRequest},
  http::StatusCode,
  FromRequest, HttpRequest, ResponseError,
};
use futures_util::future::{ready, Ready};
use percent_encoding::percent_decode_str;
use std::{
  convert::{TryFrom, TryInto},
  ops::Deref,
  path::{Component, Path, PathBuf},
  str::FromStr,
};
use thiserror::Error;

#[derive(Debug, Error, PartialEq)]
pub enum PathBufError {
  /// The segment started with the wrapped invalid character.
  #[error("The segment started with the wrapped invalid character: {}", char)]
  BadStart { char: char },
  /// The segment contained the wrapped invalid character.
  #[error("The segment contained the wrapped invalid character: {}", char)]
  BadChar { char: char },
  /// The segment ended with the wrapped invalid character.
  #[error("The segment ended with the wrapped invalid character: {}", char)]
  BadEnd { char: char },
  /// The path is not a valid UTF-8 string after percent-decoding
  #[error("The path is not a valid UTF-8 string after percent-decoding.")]
  NotValidUtf8,
  /// The path has invalid or unexpected components.
  #[error("The path has invalid or unexpected components.")]
  InvalidComponents,
}

/// Return `BadRequest` for `PathBufError`.
impl ResponseError for PathBufError {
  fn status_code(&self) -> StatusCode {
    StatusCode::BAD_REQUEST
  }
}

type Result<T, E = PathBufError> = std::result::Result<T, E>;

#[derive(Debug)]
pub struct UriPathBuf(PathBuf);

impl UriPathBuf {
  pub fn new(raw_path: &str) -> Result<Self> {
    let mut path = PathBuf::new();

    let mut segment_count = raw_path.matches('/').count() + 1;

    let decoded_path = percent_decode_str(raw_path)
      .decode_utf8()
      .map_err(|_| PathBufError::NotValidUtf8)?;

    if segment_count != decoded_path.matches('/').count() + 1 {
      return Err(PathBufError::BadChar { char: '/' });
    }

    for segment in decoded_path.split('/') {
      if segment == ".." {
        segment_count -= 1;
        path.pop();
      } else if segment.starts_with('.') {
        return Err(PathBufError::BadStart { char: '.' });
      } else if segment.starts_with('*') {
        return Err(PathBufError::BadStart { char: '*' });
      } else if segment.ends_with(':') {
        return Err(PathBufError::BadEnd { char: ':' });
      } else if segment.ends_with('>') {
        return Err(PathBufError::BadEnd { char: '>' });
      } else if segment.ends_with('<') {
        return Err(PathBufError::BadEnd { char: '<' });
      } else if segment.is_empty() {
        segment_count -= 1;
        continue;
      } else if cfg!(windows) && segment.contains('\\') {
        return Err(PathBufError::BadChar { char: '\\' });
      } else if cfg!(windows) && segment.contains(':') {
        return Err(PathBufError::BadChar { char: ':' });
      } else {
        path.push(segment)
      }
    }

    for (i, component) in path.components().enumerate() {
      if !matches!(component, Component::Normal(_)) || i >= segment_count {
        return Err(PathBufError::InvalidComponents);
      }
    }

    Ok(UriPathBuf(path))
  }
}

impl FromStr for UriPathBuf {
  type Err = PathBufError;

  fn from_str(raw_path: &str) -> Result<Self, Self::Err> {
    Self::new(raw_path)
  }
}

impl FromRequest for UriPathBuf {
  type Error = PathBufError;
  type Future = Ready<Result<Self, Self::Error>>;

  fn from_request(request: &HttpRequest, _: &mut Payload) -> Self::Future {
    ready(request.try_into())
  }
}

impl TryFrom<&HttpRequest> for UriPathBuf {
  type Error = PathBufError;

  fn try_from(request: &HttpRequest) -> Result<Self, PathBufError> {
    request.match_info().unprocessed().parse()
  }
}

impl TryFrom<&ServiceRequest> for UriPathBuf {
  type Error = PathBufError;

  fn try_from(request: &ServiceRequest) -> Result<Self, PathBufError> {
    request.match_info().unprocessed().parse()
  }
}

impl From<UriPathBuf> for PathBuf {
  fn from(path: UriPathBuf) -> PathBuf {
    path.0
  }
}

impl AsRef<Path> for UriPathBuf {
  fn as_ref(&self) -> &Path {
    self.0.deref()
  }
}

#[cfg(test)]
mod tests {
  use super::{PathBufError, UriPathBuf};
  use std::{iter::FromIterator, path::PathBuf};

  #[actix_rt::test]
  async fn test_path_buf() {
    let cases: &[(&str, Result<PathBuf, PathBufError>)] = &[
      ("/test/.tt", Err(PathBufError::BadStart { char: '.' })),
      ("/test/*tt", Err(PathBufError::BadStart { char: '*' })),
      ("/test/tt:", Err(PathBufError::BadEnd { char: ':' })),
      ("/test/tt<", Err(PathBufError::BadEnd { char: '<' })),
      ("/test/tt>", Err(PathBufError::BadEnd { char: '>' })),
      ("hello%20world", Ok(PathBuf::from_iter(vec!["hello world"]))),
      (
        "/testing%21/hello%20world",
        Ok(PathBuf::from_iter(vec!["testing!", "hello world"])),
      ),
      ("/seg1/seg2/", Ok(PathBuf::from_iter(vec!["seg1", "seg2"]))),
      ("/seg1/../seg2/", Ok(PathBuf::from_iter(vec!["seg2"]))),
      (
        "/../../../../../dev/null",
        Ok(PathBuf::from_iter(vec!["dev/null"])),
      ),
      (
        "/../../../..%2F../dev/null",
        Err(PathBufError::BadChar { char: '/' }),
      ),
    ];

    for (input, expected) in cases {
      assert_eq!(&UriPathBuf::new(input).map(|t| t.0), expected)
    }
  }
}
