use super::{
  directory::{index::IndexStrategy, Directory},
  named_ext::NamedFileExt,
  pages::render_server_page,
  path::resolve_path_within_tree,
  path_context::PathContext,
  pathbuf::UriPathBuf,
};
use actix_files::NamedFile;
use actix_service::boxed::BoxService;
use actix_web::{
  body::BoxBody,
  dev::{Service, ServiceRequest, ServiceResponse},
  error::Error as ActixError,
  guard::Guard,
  http::{header, Method, StatusCode},
  HttpResponse, ResponseError,
};
use async_recursion::async_recursion;
use futures_util::future::LocalBoxFuture;
use std::{
  convert::TryInto,
  io,
  ops::Deref,
  path::{Path, PathBuf},
  rc::Rc,
  sync::Arc,
  task::{Context, Poll},
};
use thiserror::Error;
use tokio::sync::RwLock;

/// Errors which can occur when serving static files.
#[derive(Debug, Error)]
pub enum ServiceError {
  /// Cannot render directory
  #[error("Unable to render directory without index file")]
  IsDirectory,

  #[error("Serve directory is not ready.")]
  ServeDirNotReady,

  #[error("Unable to find or open the specified file.")]
  OpenFile {
    #[source]
    source: io::Error,
  },

  #[error("Unable to canonicalize path")]
  CanonicalizePath {
    #[source]
    source: io::Error,
  },

  #[error("Method {} is not allowed", method)]
  MethodNotAllowed { method: Method },

  #[error(transparent)]
  BuildUriPath {
    #[from]
    source: super::pathbuf::PathBufError,
  },

  #[error("Unable to render listing")]
  RenderListing { source: io::Error },

  #[error("Custom path not found")]
  ServeCustomNotFoundPath,
}

/// Return `NotFound` for `ServiceError`
impl ResponseError for ServiceError {
  fn error_response(&self) -> HttpResponse {
    match self {
      ServiceError::MethodNotAllowed { .. } => HttpResponse::MethodNotAllowed()
        .append_header((header::CONTENT_TYPE, "text/plain"))
        .body("Request did not meet this resource's requirements."),
      _ => HttpResponse::new(StatusCode::NOT_FOUND),
    }
  }
}

type HttpService = BoxService<ServiceRequest, ServiceResponse, ActixError>;

struct RequestContext {
  serve_dir: PathBuf,
  path: PathBuf,
  path_context: PathContext,
}

#[derive(Clone)]
pub struct FilesService(pub(crate) Rc<FilesServiceInner>);

impl Deref for FilesService {
  type Target = FilesServiceInner;

  fn deref(&self) -> &Self::Target {
    &self.0
  }
}

pub struct FilesServiceInner {
  directory: Arc<RwLock<Option<PathBuf>>>,
  redirect_to_slash: bool,
  pub(crate) default: Option<HttpService>,
  root_path_context: Arc<PathContext>,
  path_contexts: Arc<Vec<PathContext>>,
  guards: Option<Rc<dyn Guard>>,
}

impl FilesServiceInner {
  pub(crate) fn new(
    directory: Arc<RwLock<Option<PathBuf>>>,
    redirect_to_slash: bool,
    default: Option<HttpService>,
    root_path_context: Arc<PathContext>,
    path_contexts: Arc<Vec<PathContext>>,
    guards: Option<Rc<dyn Guard>>,
  ) -> FilesServiceInner {
    FilesServiceInner {
      directory,
      redirect_to_slash,
      default,
      root_path_context,
      path_contexts,
      guards,
    }
  }

  pub fn set_default(&mut self, default: Option<HttpService>) {
    self.default = default
  }

  #[async_recursion(?Send)]
  async fn handle_err(
    &self,
    e: ServiceError,
    req: ServiceRequest,
    request_context: Option<RequestContext>,
  ) -> Result<ServiceResponse, ActixError> {
    log::debug!("Files: Failed to handle {}: {}", req.path(), e);

    if let Some(ref default) = self.default {
      log::info!("Using default handler");

      default.call(req).await
    } else {
      let error_response = e.error_response();

      match error_response.status() {
        StatusCode::NOT_FOUND => self.handle_not_found_err(req, request_context).await,
        _ => Ok(req.into_response(error_response)),
      }
    }
  }

  async fn get_request_context_for_request(
    &self,
    req: &ServiceRequest,
  ) -> Result<RequestContext, ServiceError> {
    let serve_dir = self.get_serve_dir().await?;
    let path_from_request: UriPathBuf = req.try_into()?;

    Ok(RequestContext {
      path: serve_dir.join(&path_from_request),
      serve_dir,
      path_context: self.get_path_context(&path_from_request),
    })
  }

  async fn handle_early_err(
    &self,
    e: ServiceError,
    req: ServiceRequest,
  ) -> Result<ServiceResponse, ActixError> {
    let request_context = self.get_request_context_for_request(&req).await;

    self.handle_err(e, req, request_context.ok()).await
  }

  async fn handle_not_found_err(
    &self,
    req: ServiceRequest,
    request_context: Option<RequestContext>,
  ) -> Result<ServiceResponse, ActixError> {
    if let Some(request_context) = request_context {
      log::debug!(
        "Using custom error path: {:?}",
        request_context.path_context.get_not_found_path()
      );

      // Try to use a custom 404 page, if configured.
      if let Some(not_found_path) = request_context.path_context.get_not_found_path() {
        match resolve_path_within_tree(
          &request_context.serve_dir,
          &request_context.path,
          not_found_path,
        ) {
          Ok(path) => {
            log::debug!("Custom error page: {:?}", path);

            return self
              .handle_file(
                req,
                RequestContext {
                  path,
                  ..request_context
                },
                false,
              )
              .await;
          }
          Err(err) => log::error!("Error resolving custom error page: {}", err),
        }
      }
    }

    match render_server_page("404", "Not Found", "") {
      Ok(body) => Ok(
        req.into_response(
          HttpResponse::NotFound()
            .content_type("text/html; charset=utf-8")
            .body(body),
        ),
      ),
      Err(_) => {
        self
          .handle_err(ServiceError::ServeCustomNotFoundPath, req, None)
          .await
      }
    }
  }

  async fn handle_dir(
    &self,
    req: ServiceRequest,
    request_context: RequestContext,
  ) -> Result<ServiceResponse, ActixError> {
    // Check if we should redirect the client to a normalized path ending in a
    // slash.
    if self.redirect_to_slash && !req.path().ends_with('/') {
      let redirect_to = format!("{}/", req.path());

      return Ok(
        req.into_response(
          HttpResponse::Found()
            .append_header((header::LOCATION, redirect_to))
            .body(""),
        ),
      );
    }

    let RequestContext {
      path,
      path_context,
      serve_dir,
    } = &request_context;

    return match path_context.get_index_strategy_or_default() {
      IndexStrategy::AlwaysShowListing { renderer } => {
        let dir = Directory::new(serve_dir.clone(), path.clone());
        let (http_req, payload) = req.into_parts();

        match renderer(&dir, &http_req) {
          Ok(resp) => Ok(resp),
          Err(source) => {
            let req = ServiceRequest::from_parts(http_req, payload);

            self
              .handle_err(
                ServiceError::RenderListing { source },
                req,
                Some(request_context),
              )
              .await
          }
        }
      }
      IndexStrategy::ShowListingWhenAbsent {
        renderer,
        filenames,
      } => {
        for filename in filenames {
          let path = path.join(filename);

          if path.exists() {
            return self
              .handle_file(
                req,
                RequestContext {
                  serve_dir: request_context.serve_dir,
                  path,
                  path_context: request_context.path_context,
                },
                true,
              )
              .await;
          }
        }

        let dir = Directory::new(serve_dir.clone(), path.clone());
        let (http_req, payload) = req.into_parts();

        match renderer(&dir, &http_req) {
          Ok(resp) => Ok(resp),
          Err(source) => {
            let req = ServiceRequest::from_parts(http_req, payload);

            self
              .handle_err(
                ServiceError::RenderListing { source },
                req,
                Some(request_context),
              )
              .await
          }
        }
      }
      IndexStrategy::IndexFiles { filenames } => {
        for filename in filenames {
          let path = path.join(filename);

          if path.exists() {
            return self
              .handle_file(
                req,
                RequestContext {
                  path,
                  ..request_context
                },
                true,
              )
              .await;
          }
        }

        self
          .handle_err(ServiceError::IsDirectory, req, Some(request_context))
          .await
      }
      IndexStrategy::NoIndex => {
        self
          .handle_err(ServiceError::IsDirectory, req, Some(request_context))
          .await
      }
    };
  }

  async fn handle_file(
    &self,
    req: ServiceRequest,
    request_context: RequestContext,
    preserve_request_context: bool,
  ) -> Result<ServiceResponse, ActixError> {
    let RequestContext {
      path, path_context, ..
    } = &request_context;

    match NamedFile::open(path).map_err(|source| ServiceError::OpenFile { source }) {
      Ok(named_file) => {
        let (http_req, _payload) = req.into_parts();

        let mut response = named_file
          .with_path_context(path_context)
          .into_response(&http_req);

        path_context.append_headers_to_map(response.headers_mut());

        Ok(ServiceResponse::new(http_req, response))
      }
      Err(e) => {
        self
          .handle_err(
            e,
            req,
            if preserve_request_context {
              Some(request_context)
            } else {
              None
            },
          )
          .await
      }
    }
  }

  async fn get_serve_dir(&self) -> Result<PathBuf, ServiceError> {
    let maybe_dir = self.directory.read().await.clone();

    match maybe_dir {
      Some(dir) => Ok(dir),
      None => Err(ServiceError::ServeDirNotReady),
    }
  }

  async fn get_canonical_path_from_request(
    &self,
    req: &ServiceRequest,
  ) -> Result<(PathBuf, PathBuf, UriPathBuf), ServiceError> {
    let serve_dir = self.get_serve_dir().await?;

    let path_from_request: UriPathBuf = req.try_into()?;

    // Obtain the normalized full file path to use.
    let canonical_path = serve_dir
      .join(&path_from_request)
      .canonicalize()
      .map_err(|source| ServiceError::CanonicalizePath { source })?;

    Ok((canonical_path, serve_dir, path_from_request))
  }

  fn get_path_context<P: AsRef<Path>>(&self, path_from_request: P) -> PathContext {
    let path_from_request = path_from_request.as_ref();

    for path_context in self.path_contexts.as_ref() {
      if path_context.matches_path(path_from_request) {
        log::debug!(
          "Using matched PathContext. Matcher: {:?}, Path: {}",
          path_context,
          path_from_request.display()
        );
        return self.root_path_context.merge(path_context);
      }
    }

    log::debug!(
      "Using root PathContext. Path: {}",
      path_from_request.display()
    );
    self.root_path_context.as_ref().clone()
  }
}

impl Service<ServiceRequest> for FilesService {
  type Response = ServiceResponse<BoxBody>;
  type Error = ActixError;
  type Future = LocalBoxFuture<'static, Result<Self::Response, Self::Error>>;

  fn poll_ready(&self, _: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
    Poll::Ready(Ok(()))
  }

  fn call(&self, req: ServiceRequest) -> Self::Future {
    let is_method_valid = if let Some(guard) = &self.guards {
      (**guard).check(&req.guard_ctx())
    } else {
      matches!(*req.method(), Method::HEAD | Method::GET)
    };

    let this = self.clone();

    Box::pin(async move {
      if !is_method_valid {
        return this
          .handle_early_err(
            ServiceError::MethodNotAllowed {
              method: req.method().clone(),
            },
            req,
          )
          .await;
      }

      let maybe_path = this.get_canonical_path_from_request(&req).await;
      let (path, serve_dir, path_from_request) = match maybe_path {
        Ok(path) => path,
        Err(e) => return this.handle_early_err(e, req).await,
      };

      let path_context = this.get_path_context(&path_from_request);

      let request_context = RequestContext {
        path,
        serve_dir,
        path_context,
      };

      if request_context.path.is_dir() {
        this.handle_dir(req, request_context).await
      } else {
        this.handle_file(req, request_context, true).await
      }
    })
  }
}
