#![deny(rust_2018_idioms, nonstandard_style)]
#![deny(clippy::doc_markdown, clippy::match_bool, clippy::match_same_arms)]
#![warn(future_incompatible)]

use cli::{
  args::{Opts, SubCommand},
  bundle, serve,
};
use collective::cli::ConfigurableAppOpts;
use espresso::bundle::packager;
use std::sync::Arc;
use thiserror::Error;

pub mod cli;

#[derive(Error, Debug)]
pub enum MainError {
  #[error(transparent)]
  ServeError {
    #[from]
    source: serve::ServeError,
  },
  #[error(transparent)]
  BundleError {
    #[from]
    source: packager::PackagerError,
  },
  #[error(transparent)]
  CliError {
    #[from]
    source: collective::cli::CliError,
  },
}

type Result<T, E = MainError> = std::result::Result<T, E>;

#[actix_rt::main]
async fn main() -> Result<()> {
  #[cfg(feature = "console-subscriber")]
  console_subscriber::init();

  let result = inner_main().await;

  if let Err(MainError::CliError {
    source: collective::cli::CliError::ArgParse(inner),
  }) = &result
  {
    inner.exit();
  }

  result
}

async fn inner_main() -> Result<()> {
  let (opts, config) = Opts::try_init_with_config()?;

  let config = Arc::new(config);

  match opts.subcmd {
    SubCommand::Serve => serve::serve(config).await?,
    SubCommand::Bundle(opts) => bundle::bundle(config, opts).await?,
  };

  Ok(())
}
