use crate::{
  config::{CompressionConfig, ServerConfig},
  files::{path_context::PathContext, Files},
};
use actix_http::uri::InvalidUri;
use actix_rt::Runtime;
use actix_web::{
  dev::ServerHandle,
  middleware::{self, Condition, Logger},
  App, HttpServer,
};
use collective::thread::{self, handle::ThreadHandle};
use std::{
  convert::{TryFrom, TryInto},
  path::PathBuf,
  sync::{
    mpsc::{self, RecvError, Sender},
    Arc,
  },
  thread::Thread,
};
use thiserror::Error;
use tokio::sync::RwLock;

#[derive(Debug, Error)]
pub enum ServerError {
  #[error(transparent)]
  ChannelReceive {
    #[from]
    source: RecvError,
  },
  #[error("Unable to bind")]
  Bind {
    #[source]
    source: std::io::Error,
  },
  #[error("Unable to start runtime")]
  SystemRun {
    #[source]
    source: std::io::Error,
  },
  #[error("Unable to parse redirect URI")]
  RedirectUrlParseError {
    #[from]
    source: InvalidUri,
  },
  #[error("Unable to create path context")]
  CreatePathContext {
    #[from]
    source: crate::files::path_context::PathContextError,
  },
}

pub type Result<T, E = ServerError> = std::result::Result<T, E>;

pub struct Server {
  config: ServerConfig,
  serve_dir: Arc<RwLock<Option<PathBuf>>>,
}

impl Server {
  pub fn new(config: ServerConfig, serve_dir: Arc<RwLock<Option<PathBuf>>>) -> Self {
    Server { config, serve_dir }
  }

  pub async fn spawn(
    self,
    notify_sender: Sender<Thread>,
  ) -> Result<(ServerHandle, ThreadHandle<Result<()>>)> {
    log::debug!("Starting server thread");

    let (tx, rx) = mpsc::channel();

    let server_address = self.config.address;
    let serve_dir = Arc::clone(&self.serve_dir);
    let redirect_to_slash = self.config.redirect_to_slash;

    let enable_auto_compression = matches!(
      self.config.compression,
      Some(CompressionConfig::Auto) | None
    );

    let root_path_context = Arc::new(PathContext::try_from(&self.config)?);
    let mut path_contexts: Vec<PathContext> = Vec::new();

    if let Some(path_configs) = &self.config.path_configs {
      for path_config in path_configs {
        path_contexts.push(path_config.try_into()?);
      }
    }

    let path_contexts = Arc::new(path_contexts);

    let thread_handle = thread::handle::spawn(notify_sender, move || {
      let rt = Runtime::new().unwrap();

      let srv = HttpServer::new(move || {
        let mut files_service = Files::new(
          "/",
          Arc::clone(&serve_dir),
          Arc::clone(&root_path_context),
          Arc::clone(&path_contexts),
        );

        if let Some(true) = redirect_to_slash {
          files_service = files_service.redirect_to_slash_directory();
        }

        App::new()
          .wrap(Logger::default())
          .wrap(Condition::new(
            enable_auto_compression,
            middleware::Compress::default(),
          ))
          .default_service(files_service)
      })
      .bind(server_address)
      .map_err(|source| ServerError::Bind { source })?
      .shutdown_timeout(60)
      .run();

      let _ = tx.send(srv.handle());

      rt.block_on(async { srv.await })
        .map_err(|source| ServerError::SystemRun { source })?;

      Ok(())
    });

    Ok((rx.recv()?, thread_handle))
  }
}
